/* udp-nat      Creates perconns to do a single translation of reply if
		possible.  Perconn expires after first hit.
		Translates using a scheme given by args.  */

#include <linux/types.h>
#include <linux/ip.h>
#include <linux/udp.h>
#include <linux/netdevice.h>
#include <linux/inetdevice.h>
#include <linux/netfilter.h>
#include <linux/rtnetlink.h>
#include <linux/module.h>
#include <linux/modreglist.h>

#include "ip_nat.h"
#include "ip_nat_setup.h"
#include "ip_nat_perconn.h"

#define DEBUGP printk

#define IP_NAT_UDP_TIMEOUT (60*HZ)

struct ip_nat_udp_user_port
{
	struct ip_nat_udp_user_port *next;
	u_int16_t port;
};

enum ip_nat_transtype
{
	IP_NAT_TRANS_MASQUERADE,
	IP_NAT_TRANS_NAT,
	IP_NAT_TRANS_RNAT,
	IP_NAT_TRANS_REDIRECT,
};

struct ip_nat_udp
{
	enum ip_nat_transtype type;
	/* These unused if type == IP_NAT_TRANS_MASQUERADE/REDIRECT */
	u_int32_t ipmin, ipmax;
	/* Port to map to if IP_NAT_TRANS_RNAT */
	u_int16_t mapport;	
};

struct ip_nat_address_range_udp
{
	struct ip_nat_address_range range;
	u_int16_t min_port, max_port;
};

static struct ip_nat_udp_user_port *userports;

static int check(const struct ip_nat_setup *setup, size_t extrasize)
{
	struct ip_nat_udp *info = (struct ip_nat_udp *)(setup + 1);

	if (extrasize != sizeof(struct ip_nat_udp)) {
		DEBUGP("udp_nat_check: extra size %u != %u.\n",
		       extrasize, sizeof(struct ip_nat_udp));
		return 0;
	}

	/* Must specify proto == UDP. */
	if (((struct ip_nat_pktinfo *)&setup->mask)->tuplepr.proto != 0xFF
	    || setup->info.tuplepr.proto != IPPROTO_UDP) {
		DEBUGP("udp_nat_check: mask %u, proto = %u.\n",
		       ((struct ip_nat_pktinfo *)&setup->mask)->tuplepr.proto,
		       setup->info.tuplepr.proto);
		return 0;
	}

	switch (info->type) {
	case IP_NAT_TRANS_MASQUERADE:
	case IP_NAT_TRANS_NAT:
		if (setup->direction != IP_NAT_OUTGOING) {
			DEBUGP("udp_nat_check: bad dir for masq/NAT.\n");
			return 0;
		}
		break;

	case IP_NAT_TRANS_RNAT:
	case IP_NAT_TRANS_REDIRECT:
		if (setup->direction != IP_NAT_INCOMING) {
			DEBUGP("udp_nat_check: bad dir for RNAT/REDIRECT.\n");
			return 0;
		}
		break;

	default:
		DEBUGP("udp_nat_check: bad type %u.\n", info->type);
		return 0;
	}
	return 1;
}

extern struct ip_nat_setup_type udp_nat_single;

/* Returns, refreshes timer. */
static unsigned int process(struct ip_nat_perconn *perconn,
			    struct sk_buff *skb,
			    enum ip_nat_direction dir)
{
	/* Refresh timer */
	mod_timer(&perconn->timeout, jiffies + IP_NAT_UDP_TIMEOUT);
	return NF_ACCEPT;
}

/* Hands packet to userspace. */
static unsigned int process_user(struct ip_nat_perconn *perconn,
				 struct sk_buff *skb,
				 enum ip_nat_direction dir)
{
	struct ip_nat_udp *info = (struct ip_nat_udp *)perconn->cause->private;
	struct udphdr *hdr = (struct udphdr *)((u_int32_t *)skb->nh.iph 
					       + skb->nh.iph->ihl);
	enum ip_nat_direction fromserver;

	if (info->type == IP_NAT_TRANS_MASQUERADE
	    || info->type == IP_NAT_TRANS_NAT)
		fromserver = IP_NAT_INCOMING;
	else
		fromserver = IP_NAT_OUTGOING;

	process(perconn, skb, dir);

	/* Set mark to dest port so they can have separate handlers. */ 
	if (dir == fromserver)
		skb->nfmark = hdr->source;
	else 
		skb->nfmark = hdr->dest;

	skb->nfreason = NF_SPECIAL_SET_BY_UDP_NAT;
	return NF_USERSPACE;
}

static int port_cmp(const void *arg, const void *curr)
{
	return ((struct ip_nat_udp_user_port *)curr)->port 
		== (u_int16_t)(unsigned long)arg;
}

/* Function to call if we get a match */
unsigned int setup(const struct iphdr *iph, 
		   const struct ip_nat_pktinfo *pkt,
		   enum ip_nat_direction direction,
		   const struct ip_nat_setup *cause,
		   struct ip_nat_perconn **perconn,
		   int local)
{
	struct udphdr *hdr = (struct udphdr *)((u_int32_t *)iph + iph->ihl);
	struct ip_nat_udp *info = (struct ip_nat_udp *)cause->private;
	struct ip_nat_address_range_udp range;
	ip_nat_fn *matchfn;

	DEBUGP("udp_nat setup: Doing it to %u.\n", ntohs(hdr->source));

	range.range.next = NULL;
	range.range.range_full = 0;

	/* Keep low ports mapped to low ports, and high ports mapped to
	 * high ports. */
	if (ntohs(hdr->source) < 1024) {
		range.min_port = htons(1);
		range.max_port = htons(1023);
	} else {
		range.min_port = htons(1024);
		range.max_port = htons(65535);
	}

	/* If it's a single-hit type, then expire when a match occurs. */
	if (cause->type == &udp_nat_single) matchfn = &ip_nat_delete_maybe;
	/* Send to userspace if it's in the user port list. */
	else if (modreg_find(&userports, (void *)(unsigned long)hdr->dest,
			     &port_cmp))
		matchfn = &process_user;
	/* Otherwise refresh timer. */
	else matchfn = &process;

	/* Set range. */
	switch (info->type) {
	case IP_NAT_TRANS_MASQUERADE:
		if (!ip_nat_masq_set(iph, &range.range)) {
			DEBUGP("icmp-nat: unroutable packet.\n");
			dump_iph(__FUNCTION__, iph);
			return NF_DROP;
		}
		break;

	case IP_NAT_TRANS_REDIRECT:
		if (info->mapport) 
			range.min_port = range.max_port = info->mapport;
		else range.min_port = range.max_port = hdr->dest;
		
		if (!ip_nat_interface(pkt->device, &range.range))
			return NF_DROP;
		break;

	case IP_NAT_TRANS_RNAT:
		if (info->mapport) 
			range.min_port = range.max_port = info->mapport;
		else range.min_port = range.max_port = hdr->dest;
		/* Fall thru ... */
	case IP_NAT_TRANS_NAT:
		DEBUGP("udp-nat: ip_min = %u.%u.%u.%u ip_max = %u.%u.%u.%u\n",
		       IP_PARTS(info->ipmin),
		       IP_PARTS(info->ipmax));
		range.range.min_ip = info->ipmin;
		range.range.max_ip = info->ipmax;

		break;
	}

	dump_tuplepr(__FUNCTION__, direction, &pkt->tuplepr);
	*perconn = ip_nat_alloc_perconn(cause, &pkt->tuplepr, direction,
					local,
					&range.range, 
					matchfn, 
					/* ICMP -> expire */
					&ip_nat_delete_perconn,
					IP_NAT_UDP_TIMEOUT,
					&ip_nat_delete_perconn_timeout,
					NULL, 0);
	if (!*perconn) {
		DEBUGP("udp-nat: ip_nat_alloc_perconn returned NULL.\n");
		return NF_DROP;
	}
	else DEBUGP("udp-nat: ip_nat_alloc_perconn returned ok.\n");
	return NF_ACCEPT;
}

static int print_setup(char *buffer, const struct ip_nat_setup *setup)
{
	struct ip_nat_udp *info = (struct ip_nat_udp *)setup->private;
	int len;

	switch (info->type) {
	case IP_NAT_TRANS_MASQUERADE:
		len = sprintf(buffer, " MASQ");
		break;

	case IP_NAT_TRANS_REDIRECT:
		len = sprintf(buffer, " REDIR");
		if (info->mapport)
			len += sprintf(buffer+len, " to port %hu",
				       info->mapport);
		break;

	case IP_NAT_TRANS_NAT:
		len = sprintf(buffer, " NAT %u.%u.%u.%u",
			      IP_PARTS_NATIVE(info->ipmin));
		if (info->ipmin != info->ipmax)
			len += sprintf(buffer + len, "-%u.%u.%u.%u",
				       IP_PARTS_NATIVE(info->ipmax));
		break;

	default:
		len = sprintf(buffer, " RNAT %u.%u.%u.%u", 
			      IP_PARTS_NATIVE(info->ipmin));
		if (info->ipmin != info->ipmax)
			len += sprintf(buffer + len, "-%u.%u.%u.%u",
				       IP_PARTS_NATIVE(info->ipmax));
		if (info->mapport)
			len += sprintf(buffer + len, " to port %hu",
				       info->mapport);
		break;
	}
	return len;
}

static int do_udp_ctl(int cmd, void *user, unsigned int len)
{
	struct ip_nat_udp_user_port *portinfo;

	switch (cmd) {
	case IP_NAT_UDP_ADD_USER:
		portinfo = kmalloc(sizeof(*portinfo), GFP_USER);
		if (!portinfo)
			return -ENOMEM;
		else {
			portinfo->port = (u_int16_t)(unsigned long)user;
			modreg_prepend(&userports, portinfo);
		}
		break;

	case IP_NAT_UDP_DELETE_USER:
		portinfo = modreg_find(&userports, user, &port_cmp);

		if (!portinfo)
			return -ENOENT;
		else {
			modreg_delete(&userports, portinfo);
			kfree(portinfo);
		}
	}

	return 0;
}

static struct nf_setsockopt_ops ip_nat_udp_sockopts 
= { NULL, PF_INET, IP_NAT_UDP_ADD_USER, IP_NAT_UDP_DELETE_USER, do_udp_ctl };

/* Print out the private part of the perconn (if any). */
static int print_perconn(char *buffer, const struct ip_nat_perconn *perconn)
{
	if (perconn->postmangle == &process_user)
		return sprintf(buffer, " USER");

	return 0;
}

static struct ip_nat_setup_type udp_nat
= { NULL, "udp-nat", 0, &__this_module, check, setup, print_setup, 
    print_perconn };
static struct ip_nat_setup_type udp_nat_single
= { NULL, "udp-nat-single", 0, &__this_module, check, setup, print_setup,
    print_perconn };

static int init_or_cleanup(int init)
{
	int ret = 0;
	if (!init) goto cleanup;

	ret = ip_nat_register_setup_type(&udp_nat);
	if (ret < 0) goto cleanup_nothing;

	ret = ip_nat_register_setup_type(&udp_nat_single);
	if (ret < 0) goto cleanup_udp;

	ret = nf_register_sockopt(&ip_nat_udp_sockopts);
	if (ret < 0) goto cleanup_udpudpsingle;

	return ret;

 cleanup:
	nf_unregister_sockopt(&ip_nat_udp_sockopts);
 cleanup_udpudpsingle:
	ip_nat_unregister_setup_type(&udp_nat_single);
 cleanup_udp:
	ip_nat_unregister_setup_type(&udp_nat);
 cleanup_nothing:

	return ret;
}

int init_module(void)
{
	return init_or_cleanup(1);
}

void cleanup_module()
{
	init_or_cleanup(0);
}
