/* FTP redirect/RNAT: Alter PASV responses (from server, ie. direction
 * IP_NAT_OUTGOING), and create RNAT for that new incoming connection.
 *
 * FTP masquerade/NAT: Alter PORT commands (from client, ie. direction
 * IP_NAT_OUTGOING), and create NAT for that connection.  
 *
 * FIXME: We really need a way of creating a perconn; we need the port
 * bound before we tell the other end what it is.  This is also
 * required for (eg.) RealAudio streams.
 */
#include <sys/types.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <fcntl.h>
#include <net/if.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <sys/ioctl.h>
#include <linux/netfilter_dev.h>
#include <linux/netfilter_ipv4.h>
#include <arpa/inet.h>
#include "ip_nat.h"
#include "ip_nat_setup.h"

#define DEBUGP printf

#define IP_PARTS_NATIVE(n)			\
(unsigned int)((n)>>24)&0xFF,			\
(unsigned int)((n)>>16)&0xFF,			\
(unsigned int)((n)>>8)&0xFF,			\
(unsigned int)((n)&0xFF)

#define IP_PARTS(n) IP_PARTS_NATIVE(ntohl(n))

enum ip_nat_transtype
{
	IP_NAT_TRANS_MASQUERADE,
	IP_NAT_TRANS_NAT,
	IP_NAT_TRANS_RNAT,
	IP_NAT_TRANS_REDIRECT,
};

struct ip_nat_tcp_udp
{
	enum ip_nat_transtype type;
	/* These unused if type == IP_NAT_TRANS_MASQUERADE/REDIRECT */
	u_int32_t ipmin, ipmax;
	/* Port to map to if IP_NAT_TRANS_RNAT/REDIRECT */
	u_int16_t mapport;
};

struct rnat
{
	struct ip_nat_setup_user base;
	struct ip_nat_tcp_udp tcpudp;
};

/* Need one per connection. */
struct conn
{
	struct conn *next;
	u_int32_t src, dst;
	u_int16_t srcpt, dstpt;

	/* RNAT: client messages come off NF_IP_LOCAL_OUT or NF_IP_PRE_ROUTING
	   NAT: client messages come off NF_IP_POST_ROUTING.
	*/
	int nat;

	/* Fortunately, we never have more than one offset in
           progress, because of they way FTP works. */
	struct {
		u_int32_t pos;
		int before;
		int after;
	} offset;

	int expecting_pasv_response;

	/* The return path (if any) we have created at the moment */
	struct rnat *rnat;
};

static struct conn *list;

static __inline int before(u_int32_t seq1, u_int32_t seq2)
{
        return (int32_t)(seq1-seq2) < 0;
}

static __inline int after(u_int32_t seq1, u_int32_t seq2)
{
	return (int32_t)(seq2-seq1) < 0;
}

/* Given this protocol, srcip, dstip, srcpt and dstpt (host order),
   and whether it's NAT or RNAT, return the replacement ip/port.
   Returns bool. */
static int getorig(u_int8_t proto, u_int32_t srcip, u_int32_t dstip, 
		   u_int16_t srcpt, u_int16_t dstpt, int nat,
		   u_int32_t *ip, u_int16_t *pt)
{
	/* Look through /proc for connection entry. */
	FILE *fp = fopen("/proc/net/ip_nat_connections", "r");
	char line[200];

	if (!fp)
		return 0;

	/* We read these in as ascii, so they'll be in host order. */
	srcpt = ntohs(srcpt);
	dstpt = ntohs(dstpt);
	
	while (fgets(line, sizeof(line), fp)) {
		char *sip, *dip, *oldip, *spt, *dpt, *oldpt;

		if (proto == IPPROTO_TCP && strcmp(line, "tcp_nat ") != 0)
			continue;
		else if (proto == IPPROTO_UDP && strcmp(line, "udp_nat ") != 0)
			continue;

		if (strcmp(line + strlen(line) - 6, " USER\n") != 0)
			continue;

		if (nat) {
			char *in_ptr = strstr(line, " out:") + 5;
			/* Old IP is dest address. */
			oldip = strstr(in_ptr, " dst=") + 5;
			sip = strstr(in_ptr, " src=") + 5;
			dip = strstr(in_ptr, " new ") + 9;
			
			oldpt = strstr(in_ptr, " dstpt=") + 7;
			spt = strstr(in_ptr, " srcpt=") + 7;
			dpt = strstr(in_ptr, " port=") + 6;
		} else {
			char *in_ptr = strstr(line, " in:") + 4;
			/* Old IP is source address. */
			oldip = strstr(in_ptr, " src=") + 5;
			sip = strstr(in_ptr, " new ") + 9;
			dip = strstr(in_ptr, " dst=") + 5;

			oldpt = strstr(in_ptr, " srcpt=") + 7;
			spt = strstr(in_ptr, " port=") + 6;
			dpt = strstr(in_ptr, " dstpt=") + 7;
		}

		if (inet_addr(sip) == srcip
		    && inet_addr(dip) == dstip
		    && atoi(dpt) == dstpt
		    && atoi(spt) == srcpt) {
			*ip = inet_addr(oldip);
			*pt = htons(atoi(oldpt));
			fclose(fp);
			return 1;
		}
	}

	fclose(fp);
	return 0;
}

/* We need to set up a port at the original destination IP which
   gets NAT'd to the ip/port given, and alter the packet. */
static void open_rnat(int handle, 
		      struct conn *conn,
		      struct nfdev_head *head,
		      char *start, char *end, int rest,
		      u_int32_t ip, u_int16_t port, u_int32_t syn)
{
	u_int32_t orig_ip;
	u_int16_t orig_port;
	int lendiff;
	char buffer[3 * 6 + 5 + 1];

	/* Tell it to go to original IP's ftp-data port.  Match
	   connections coming in from that source IP. */
	if (!getorig(IPPROTO_TCP, conn->src, conn->dst, 
		     conn->srcpt, conn->dstpt, 0, 
		     &orig_ip, &orig_port)) {
		DEBUGP("Strange: can't find rnat connection.\n");
		return;
	}

	sprintf(buffer, "%u,%u,%u,%u,%u,%u", IP_PARTS(orig_ip), 0, 20);
	lendiff = strlen(buffer) - (end - 1 - start);
	/* Make room for new PASV response details */
	memmove(end, end + lendiff, rest);
	memcpy(start, buffer, strlen(buffer));
	conn->offset.before = conn->offset.after;
	conn->offset.after += lendiff;
	conn->offset.pos = syn;

	head->pktlen = (int)head->pktlen + lendiff;

	if (conn->rnat) {
		/* Delete old rnat. */
		if (setsockopt(handle, IPPROTO_IP, IP_NAT_SO_DELETE, 
			       conn->rnat, sizeof(*conn->rnat)) < 0)
			DEBUGP("Strange: deleting old RNAT: %s.\n", 
			       strerror(errno));
		free(conn->rnat);
		conn->rnat = NULL;
	}
	conn->rnat = malloc(sizeof(*conn->rnat));

	/* Packet fixed.  Now open RNAT. */
	memset(conn->rnat, 0, sizeof(*conn->rnat));
	strcpy(conn->rnat->base.typename, "tcp-nat");
	conn->rnat->base.direction = IP_NAT_INCOMING;
	conn->rnat->base.tupleproto = ((struct ip_nat_tuple_proto)
				       { IPPROTO_TCP, 
					 { { conn->src, { { 0 } } }, 
					   { orig_ip, { { htons(20) }}}}});
	conn->rnat->base.mask.tuplepr.proto = 0xFFFFFFFF;
	conn->rnat->base.mask.tuplepr.tuple.src.ip = 0xFFFFFFFF;
	conn->rnat->base.mask.tuplepr.tuple.dst.ip = 0xFFFFFFFF;
	conn->rnat->base.mask.tuplepr.tuple.dst.u.tcp.port = 0xFFFF;

	conn->rnat->tcpudp.type = IP_NAT_TRANS_RNAT;
	conn->rnat->tcpudp.ipmin = conn->rnat->tcpudp.ipmax = ip;
	conn->rnat->tcpudp.mapport = port;

	if (setsockopt(handle, IPPROTO_IP, IP_NAT_SO_INSERT, conn->rnat,
		       sizeof(*conn->rnat)) < 0) 
		DEBUGP("Can't insert new RNAT.\n");
}

/* Register to handle a port; returns -1 or netfilter_dev file handle.
 * If handle is -1, opens /dev/nf_ipv4, otherwise tries to attach that
 * protocol to the handle given.  */
static int ftp_register(void)
{
	int handle;
	struct nfdev_condition cond;

	/* Setup handle to receive packets first. */
	cond = ((struct nfdev_condition){ 0xFFFFFFFF, 0, 0 });
	handle = open("/dev/netfilter_ipv4", O_RDWR);

	/* Delete original condition. */
	if (handle < 0
	    || ioctl(handle, NFDIOSDELCOND, &cond) < 0) {
		close(handle);
		return -1;
	}
	/* FIXME: blow off any captured packets. */

	/* We need PRE_ROUTING, LOCAL_OUT POST_ROUTING */
	cond = ((struct nfdev_condition){ 
		(1<<NF_IP_PRE_ROUTING)
		| (1<<NF_IP_LOCAL_OUT)
		| (1<<NF_IP_POST_ROUTING), 
		htons(21), 
		NF_SPECIAL_SET_BY_TCP_NAT });

	if (ioctl(handle, NFDIOSADDCOND, &cond) < 0) {
		close(handle);
		return -1;
	}
	return handle;
}

/* FIXME:  Handle FINs & reset on SYNs */
static struct conn *find_or_add_conn(struct iphdr *iph, struct tcphdr **tcph, 
				     int hook)
{
	struct conn **i;

	*tcph = (void *)((u_int32_t *)iph + iph->ihl);

	for (i = &list; *i; i = &(*i)->next) {
		if ((*i)->src == iph->saddr
		    && (*i)->dst == iph->daddr
		    && (*i)->srcpt == (*tcph)->source
		    && (*i)->dstpt == (*tcph)->dest) 
			return *i;
	}
	*i = malloc(sizeof(struct conn));
	**i = ((struct conn){ NULL, iph->saddr, iph->daddr, 
			      (*tcph)->source, (*tcph)->dest, 
			      /* First message is from client. */
			      hook == NF_IP_POST_ROUTING ? 1 : 0,
			      { 0, 0, 0 }, 0, NULL });
	return *i;
}

static void do_pasv(int handle, struct nfdev_head *head, 
		    struct conn *conn, char *data, unsigned int dlen,
		    u_int32_t syn)
{
	char *start, *end;
	/* From RFC 959:
	 * 227 Entering Passive Mode (h1,h2,h3,h4,p1,p2). 
	 */
	if (dlen > 4) {
		if (memcmp(data, "227 ", 4) == 0) {
			unsigned int h1,h2,h3,h4,p1,p2;
			start = memchr(data, '(', dlen);
			end = memchr(data, ')', dlen);
			if (!start || !end) {
				((char *)data)[dlen] = '\0';
				DEBUGP("Bad string `%s'\n", 
				       data);
			} else {
				*end = '\0';
				if (sscanf(start+1, 
					   "%u,%u,%u,%u,%u,%u",
					   &h1, &h2, &h3, &h4, &p1, &p2) 
				    != 6) {
					DEBUGP("Bad ip `%s'\n", start);
				} else {
					open_rnat(handle, conn, head, data, 
						  end,
						  (data + dlen) - end,
						  (h1 << 24)
						  | (h2 << 16)
						  | (h3 << 8)
						  | h4,
						  (p1 << 8)|p2,
						  syn);
				}
				*end = ')';
			}
		/* Valid errors: 500, 501, 502, 421, 530 */
		} 
		else if (memcmp(data, "500 ", 4) == 0
			 || memcmp(data, "501 ", 4) == 0
			 || memcmp(data, "502 ", 4) == 0
			 || memcmp(data, "421 ", 4) == 0
			 || memcmp(data, "530 ", 4) == 0) {
			((char *)data)[dlen] = '\0';
			DEBUGP("Normal error: `%s'\n",  data);
		} else {
			((char *)data)[dlen] = '\0';
			DEBUGP("Bad response: `%s'\n", data);
		}
	} else {
		((char *)data)[dlen] = '\0';
		DEBUGP("Short response (%u chars): `%s'\n", dlen, data);
	}
}

static void do_port(int handle, struct nfdev_head *head, 
		    struct conn *conn, char *data, unsigned int dlen,
		    u_int32_t syn)
{
	char *end = memchr(data, '\r', dlen);
	if (end) {
		unsigned int h1,h2,h3,h4,p1,p2;

		*end = '\0';
		if (sscanf(data + sizeof("port ")-1, "%u,%u,%u,%u,%u,%u",
			   &h1, &h2, &h3, &h4, &p1, &p2) 
		    != 6) {
			DEBUGP("Bad ip `%s'\n", data);
		} else {
			open_rnat(handle, conn, head, data+sizeof("port ")-1,
				  end, (data + dlen) - end,
				  (h1 << 24) | (h2 << 16) | (h3 << 8) | h4,
				  (p1 << 8) | p2,
				  syn);
		}
		*end = '\r';
	}
}

int main()
{
	struct {
		struct nfdev_head head;
		char data[65536];
	} s;
	int handle = ftp_register();

	if (handle < 0) {
		perror("Registering interest");
		return 1;
	}

	while (read(handle, &s, sizeof(s)) > 0) {
		/* Find connection. */
		struct tcphdr *tcph;
		struct conn *conn = find_or_add_conn((struct iphdr *)s.data,
						     &tcph, s.head.hook);
		void *data = (u_int32_t *)tcph + tcph->doff;
		unsigned int dlen = s.head.pktlen - ((char *)data - s.data);
		struct nfdev_verdict verdict;

		if (!conn->nat) {
			/* RNAT */
			if (s.head.hook == NF_IP_POST_ROUTING) {
				/* Server messages. */
				if (conn->expecting_pasv_response) {
					do_pasv(handle, &s.head, conn,
						s.data, dlen,
						ntohl(tcph->seq));
					conn->expecting_pasv_response = 0;
				}
				if (after(tcph->seq, conn->offset.pos))
					tcph->seq = htonl(ntohl(tcph->seq)
							  +conn->offset.after);
				else 
					tcph->seq =htonl(ntohl(tcph->seq)
							 +conn->offset.before);
			} else {
				/* Client messages. */
				if (dlen == sizeof("PASV\r\n")-1
				    && (memcmp(data, "PASV\r\n", dlen) == 0
					|| memcmp(data, "pasv\r\n", dlen)==0))
					conn->expecting_pasv_response = 1;

				if (after(tcph->ack_seq, conn->offset.pos))
					tcph->ack_seq
						= htonl(ntohl(tcph->ack_seq)
							+ conn->offset.after);
				else 
					tcph->ack_seq 
						= htonl(ntohl(tcph->ack_seq)
							+ conn->offset.before);
			}
		} else {
			/* NAT */
			if (s.head.hook == NF_IP_POST_ROUTING) {
				/* Client messages. */
				if (dlen > sizeof("port ")
				    && (memcmp(data, "port ", 
					       sizeof("port ")-1) == 0
					|| memcmp(data, "PORT ", 
						  sizeof("PORT ")-1) == 0))
					do_port(handle, &s.head, conn,
						s.data, dlen,
						ntohl(tcph->seq));

				if (after(tcph->seq, conn->offset.pos))
					tcph->seq = htonl(ntohl(tcph->seq)
							  +conn->offset.after);
				else 
					tcph->seq =htonl(ntohl(tcph->seq)
							 +conn->offset.before);
			} else {
				/* Server messages. */
				if (after(tcph->ack_seq, conn->offset.pos))
					tcph->ack_seq
						= htonl(ntohl(tcph->ack_seq)
							+ conn->offset.after);
				else 
					tcph->ack_seq 
						= htonl(ntohl(tcph->ack_seq)
							+ conn->offset.before);
			}
		}

		/* Write new packet back */
		lseek(handle, sizeof(s.head), SEEK_SET);
		write(handle, s.data, s.head.pktlen);
		verdict = ((struct nfdev_verdict)
			   { NF_ACCEPT, s.head.pktlen, s.head.mark, 
			     s.head.reason });
		if (ioctl(handle, NFDIOSVERDICT, &verdict) != 0)
			DEBUGP("Problem with NFDIOSVERDICT: %s\n",
			       strerror(errno));
	}
	return 0;
}

