/* Quake servers?  AFAIK, they don't send back any port
   information, so only mangle the client's stuff. */

/* Details and some code stolen from 2.2 masq code, by Harald Hoyer. */
#include <sys/types.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <fcntl.h>
#include <net/if.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/udp.h>
#include <sys/ioctl.h>
#include <linux/netfilter_dev.h>
#include <linux/netfilter_ipv4.h>
#include <arpa/inet.h>
#include "ip_nat.h"
#include "ip_nat_setup.h"

#define DEBUGP printf

#define IP_PARTS_NATIVE(n)			\
(unsigned int)((n)>>24)&0xFF,			\
(unsigned int)((n)>>16)&0xFF,			\
(unsigned int)((n)>>8)&0xFF,			\
(unsigned int)((n)&0xFF)

#define IP_PARTS(n) IP_PARTS_NATIVE(ntohl(n))

/* Quake header. */
struct quake
{ 
	struct {
		unsigned char type[2];     /* (Little Endian) Type . */
		unsigned char length[2];   /* (Little Endian) Length,
					      header included. */
		unsigned char code;
	} common;
	union {
		unsigned char message[sizeof("QUAKE")+1]; /* code = 0x01 */
		unsigned char port[4];			  /* code = 0x81 */
	} u;
};

/* Need one per connection. */
struct conn
{
	struct conn *next;
	u_int32_t src, dst;
	u_int16_t srcpt, dstpt;

	/* Don't shoot until you see the whites of their eyes. */
	int version_seen;
};

/* Register to handle a port; returns -1 or netfilter_dev file handle.
 * If handle is -1, opens /dev/nf_ipv4, otherwise tries to attach that
 * protocol to the handle given.  */
static int quake_register(void)
{
	int handle;
	struct nfdev_condition cond;

	/* Setup handle to receive packets first. */
	cond = ((struct nfdev_condition){ 0xFFFFFFFF, 0, 0 });
	handle = open("/dev/netfilter_ipv4", O_RDWR);

	/* Delete original condition. */
	if (handle < 0
	    || ioctl(handle, NFDIOSDELCOND, &cond) < 0) {
		close(handle);
		return -1;
	}
	/* FIXME: blow off any captured packets. */

	/* We need POST_ROUTING only. */
	cond = ((struct nfdev_condition){ 
		1<<NF_IP_POST_ROUTING, 
		htons(26000), 
		NF_SPECIAL_SET_BY_UDP_NAT });

	if (ioctl(handle, NFDIOSADDCOND, &cond) < 0) {
		close(handle);
		return -1;
	}

	cond = ((struct nfdev_condition){ 
		1<<NF_IP_POST_ROUTING, 
		htons(27000), 
		NF_SPECIAL_SET_BY_UDP_NAT });

	if (ioctl(handle, NFDIOSADDCOND, &cond) < 0) {
		close(handle);
		return -1;
	}
	return handle;
}

static struct conn *list;
static struct conn *find_or_add_conn(struct iphdr *iph, struct udphdr **udph, 
				     int hook)
{
	struct conn **i;

	*udph = (void *)((u_int32_t *)iph + iph->ihl);

	for (i = &list; *i; i = &(*i)->next) {
		if ((*i)->src == iph->saddr
		    && (*i)->dst == iph->daddr
		    && (*i)->srcpt == (*udph)->source
		    && (*i)->dstpt == (*udph)->dest) 
			return *i;
	}
	*i = malloc(sizeof(struct conn));
	**i = ((struct conn){ NULL, iph->saddr, iph->daddr, 
			      (*udph)->source, (*udph)->dest, 0 });
	return *i;
}

/* Given this protocol, srcip, dstip, srcpt and dstpt (host order),
   and whether it's NAT or RNAT, return the replacement ip/port.
   Returns bool. */
static int getorig(u_int8_t proto, u_int32_t srcip, u_int32_t dstip, 
		   u_int16_t srcpt, u_int16_t dstpt, int nat,
		   u_int32_t *ip, u_int16_t *pt)
{
	/* Look through /proc for connection entry. */
	FILE *fp = fopen("/proc/net/ip_nat_connections", "r");
	char line[200];

	if (!fp)
		return 0;

	/* We read these in as ascii, so they'll be in host order. */
	srcpt = ntohs(srcpt);
	dstpt = ntohs(dstpt);
	
	while (fgets(line, sizeof(line), fp)) {
		char *sip, *dip, *oldip, *spt, *dpt, *oldpt;

		if (proto == IPPROTO_TCP && strcmp(line, "tcp_nat ") != 0)
			continue;
		else if (proto == IPPROTO_UDP && strcmp(line, "udp_nat ") != 0)
			continue;

		if (strcmp(line + strlen(line) - 6, " USER\n") != 0)
			continue;

		if (nat) {
			char *in_ptr = strstr(line, " out:") + 5;
			/* Old IP is dest address. */
			oldip = strstr(in_ptr, " dst=") + 5;
			sip = strstr(in_ptr, " src=") + 5;
			dip = strstr(in_ptr, " new ") + 9;
			
			oldpt = strstr(in_ptr, " dstpt=") + 7;
			spt = strstr(in_ptr, " srcpt=") + 7;
			dpt = strstr(in_ptr, " port=") + 6;
		} else {
			char *in_ptr = strstr(line, " in:") + 4;
			/* Old IP is source address. */
			oldip = strstr(in_ptr, " src=") + 5;
			sip = strstr(in_ptr, " new ") + 9;
			dip = strstr(in_ptr, " dst=") + 5;

			oldpt = strstr(in_ptr, " srcpt=") + 7;
			spt = strstr(in_ptr, " port=") + 6;
			dpt = strstr(in_ptr, " dstpt=") + 7;
		}

		if (inet_addr(sip) == srcip
		    && inet_addr(dip) == dstip
		    && atoi(dpt) == dstpt
		    && atoi(spt) == srcpt) {
			*ip = inet_addr(oldip);
			*pt = htons(atoi(oldpt));
			fclose(fp);
			return 1;
		}
	}

	fclose(fp);
	return 0;
}

int main()
{
	struct {
		struct nfdev_head head;
		char data[65536];
	} s;
	int handle = quake_register();

	if (handle < 0) {
		perror("Registering interest");
		return 1;
	}

	while (read(handle, &s, sizeof(s)) > 0) {
		/* Find connection. */
		struct udphdr *udph;
		struct conn *conn = find_or_add_conn((struct iphdr *)s.data,
						     &udph, s.head.hook);
		struct quake *data = (struct quake *)(udph+1);
		unsigned int dlen = s.head.pktlen - ((char *)data - s.data);
		struct nfdev_verdict verdict;

		/* If it's an admin message... */
		if (dlen > sizeof(data->common)
		    && (data->common.type[1] & 0x80)) {
			if (data->common.code == 0x01
			    && dlen >= (sizeof(data->common)
					+ sizeof(data->u.message))
			    && memcmp(data->u.message, "QUAKE\0\3", 
				      sizeof(data->u.message)) == 0)
				/* Cool.  Saw version registration. */
				conn->version_seen = 1;
			else if (data->common.code == 0x81 
				 && conn->version_seen) {
				/* Replace port number. */
				u_int32_t ip;
				u_int16_t port;

				getorig(IPPROTO_UDP, conn->src, conn->dst,
					conn->srcpt, conn->dstpt, 1, 
					&ip, &port);

				data->u.port[0] = (ntohs(port) & 0xFF);
				data->u.port[1] = ((ntohs(port) >> 8) & 0xFF);
			}
			/* Write new packet back */
			lseek(handle, sizeof(s.head), SEEK_SET);
			write(handle, s.data, s.head.pktlen);
		}

		verdict = ((struct nfdev_verdict)
			   { NF_ACCEPT, s.head.pktlen, s.head.mark, 
			     s.head.reason });
		if (ioctl(handle, NFDIOSVERDICT, &verdict) != 0)
			DEBUGP("Problem with NFDIOSVERDICT: %s\n",
			       strerror(errno));
	}
	return 0;
}


