#ifndef _IP_NAT_H
#define _IP_NAT_H

struct sk_buff;
struct device;
struct iphdr;

#ifdef DEBUG
#define DEBUGP printk
#else
#define DEBUGP(format, args...)
#endif

/* The protocol-specific parts of the tuple which we can manupulate. */
union ip_nat_proto_manip 
{
	struct {
		u_int16_t port;
	} tcp;
	struct {
		u_int16_t port;
	} udp;
	struct {
		u_int16_t id;
	} icmp;
	/* Add other protocols here. */

	u_int32_t all[0];
};

/* These are the parts of the tuple we can manipulate (for
 * outgoing tuple allocation) */
struct ip_nat_tuple_manip 
{
	u_int32_t ip;
	union ip_nat_proto_manip u;
};

/* This contains the information to distinguish a connection. */
struct ip_nat_tuple
{
	struct ip_nat_tuple_manip src;

	/* These are the parts of the tuple which are fixed for
	 * outgoing tuple allocation. */
	struct {
		u_int32_t ip;
		union {
			struct {
				u_int16_t port;
			} tcp;
			struct {
				u_int16_t port;
			} udp;
			struct {
				u_int8_t type, code;
			} icmp;
			/* Add other protocols here. */

			u_int32_t all[0];
		} u;
	} dst;
};

struct ip_nat_tuple_proto
{
	u_int32_t proto;
	struct ip_nat_tuple tuple;
};

enum ip_nat_direction
{
	IP_NAT_INCOMING,
	IP_NAT_OUTGOING
};

struct ip_nat_pktinfo
{
	const struct device *device;
	struct ip_nat_tuple_proto tuplepr;
	enum ip_nat_direction dir;
};

/* Single range specification. */
struct ip_nat_range
{
	/* Inclusive: network order. */
	u_int32_t min_ip, max_ip;

	/* Set to 1 if this is a protocol-specific range. */
	int proto_specific;

	union ip_nat_proto_manip min, max;
};

/* FIXME: When do we clear the full flags? --RR */

/* A range consists of a linked list of ip_nat_range */
struct ip_nat_multi_range
{
	struct ip_nat_multi_range *next;
	/* Set to 0.  alloc_perconn sets this to 1 if it was
           preferred, but that protocol's unique_tuple() failed. */
	u_int32_t full;

	struct ip_nat_range range;
};

/* Calculates relative checksums. */
extern u_int16_t ip_nat_cheat_check(u_int32_t oldvalinv, u_int32_t newval,
				    u_int16_t oldcheck);

/* Check for tuple existence. */
extern int ip_nat_used_tuple(const struct ip_nat_tuple_proto *tuplepr);

#ifdef __KERNEL__
#define IP_PARTS_NATIVE(n)			\
(unsigned int)((n)>>24)&0xFF,			\
(unsigned int)((n)>>16)&0xFF,			\
(unsigned int)((n)>>8)&0xFF,			\
(unsigned int)((n)&0xFF)

#define IP_PARTS(n) IP_PARTS_NATIVE(ntohl(n))

#define IP_NAT_TUPLE_PROT_INIT(t,_src,_dst,pr)		\
do {							\
	memset((t), 0, sizeof(*(t)));			\
	(t)->proto = (pr);				\
	(t)->tuple.src.ip = (_src);			\
	(t)->tuple.dst.ip = (_dst);			\
} while(0)

static inline void dump_tuplepr(const char *fn,
				enum ip_nat_direction dir, 
				const struct ip_nat_tuple_proto *tuplepr)
{
#ifdef DEBUG
	printk("%s: %s proto=%u src=%u.%u.%u.%u id=%u "
	       "dst=%u.%u.%u.%u type=%u code=%u\n",
	       fn, dir == IP_NAT_OUTGOING ? "OUT" : "IN", 
	       tuplepr->proto, IP_PARTS(tuplepr->tuple.src.ip),
	       ntohs(tuplepr->tuple.src.u.icmp.id), 
	       IP_PARTS(tuplepr->tuple.dst.ip),
	       tuplepr->tuple.dst.u.icmp.type,
	       tuplepr->tuple.dst.u.icmp.code);
#endif
}

static inline void dump_iph(const char *fn, const struct iphdr *iph)
{
#ifdef DEBUG
	printk("%s: proto=%u src=%u.%u.%u.%u dst=%u.%u.%u.%u\n",
	       fn, iph->protocol, IP_PARTS(iph->saddr), IP_PARTS(iph->daddr));
#endif
}

static inline void dump_manip(const char *fn,
			      enum ip_nat_direction dir, 
			      const struct ip_nat_tuple_manip *manip)
{
#ifdef DEBUG
	printk("%s manip: %s=%u.%u.%u.%u id=%u\n",
	       fn, dir == IP_NAT_OUTGOING ? "src" : "dst",
	       IP_PARTS(manip->ip), manip->u.icmp.id);
#endif
}
#endif
#endif /*_IP_NAT_H*/
