/* Everything to do with perconn manipulation, particularly allocating
   them. */

/* The amount of total bytes required in a perconn. */
static size_t perconn_size = sizeof(struct ip_nat_perconn);

/* Outgoing is simple: compare with outgoing connection */
static struct ip_nat_perconn *
find_perconn(const struct ip_nat_tuple_proto *tuplepr, 
	     enum ip_nat_direction dir,
	     long now,
	     int unexpired_only)
{
	const struct ip_nat_perconn *ret;

	dump_tuplepr(__FUNCTION__, dir, tuplepr);

	for (ret = *hash_by_tuple(tuplepr, dir);
	     ret; 
	     ret = ret->by_tuple[dir]) {
		dump_tuplepr(__FUNCTION__ "looking at", dir, 
			     &ret->dir[dir].connection);
		if (memcmp(tuplepr, &ret->dir[dir].connection, 
			   sizeof(*tuplepr)) == 0) {
			if (!unexpired_only || !ret->dir[dir].expired)
				break;
		}
	}
	return (struct ip_nat_perconn *)ret;
}

void ip_nat_delete_perconn_timeout(struct ip_nat_perconn *perconn)
{
	struct ip_nat_perconn **list;
	const struct ip_nat_tuple_proto *tpr
		= &perconn->dir[IP_NAT_OUTGOING].connection;

	/* Kill timer: interrupts are blocked, so it's no race. */
	if (perconn->timerfn) {
		DEBUGP("ip_nat_delete: del_timer %p.\n", &perconn->timeout);
		del_timer(&perconn->timeout);
	}

	/* Remove from the hashes */
	list = hash_by_tuple(&perconn->dir[IP_NAT_OUTGOING].connection,
			     IP_NAT_OUTGOING);
	while (*list != perconn) 
		list = &(*list)->by_tuple[IP_NAT_OUTGOING];
	*list = perconn->by_tuple[IP_NAT_OUTGOING];

	list = hash_by_tuple(&perconn->dir[IP_NAT_INCOMING].connection,
			     IP_NAT_INCOMING);
	while (*list != perconn) 
		list = &(*list)->by_tuple[IP_NAT_INCOMING];
	*list = perconn->by_tuple[IP_NAT_INCOMING];

	list = hash_by_ips_proto(perconn->dir[IP_NAT_OUTGOING].manip.ip, 
				 tpr->tuple.dst.ip, tpr->proto);

	while (*list != perconn) 
		list = &(*list)->by_ips_proto;
	*list = perconn->by_ips_proto;

	list = hash_by_source(tpr);
	while (*list != perconn) 
		list = &(*list)->by_source;
	*list = perconn->by_source;

	/* If it has a cause (ie. not null perconn), notify bindings
	 * of destruction */
	if (perconn->cause) {
		unsigned int i;

		for (i = 0; perconn->cause->type[i]; i++) 
			perconn->cause->type[i]->destroy(perconn);
	}

	kfree(perconn);
}

/* Never fails, returns NF_ACCEPT.  Suitable for use as ip_nat_fn. 
 * You need lock to use this. */
unsigned int ip_nat_delete_perconn(struct ip_nat_perconn *perconn,
				   struct sk_buff *skb,
				   enum ip_nat_direction dir)
{
	ip_nat_delete_perconn_timeout(perconn);
	return NF_ACCEPT;
}

unsigned int ip_nat_delete_maybe(struct ip_nat_perconn *perconn,
				 struct sk_buff *skb,
				 enum ip_nat_direction dir)
{
	/* Expire this direction.  If both expired, delete. */
	perconn->dir[dir].expired = 1;
	if (perconn->dir[!dir].expired) 
		ip_nat_delete_perconn(perconn, skb, dir);

	return NF_ACCEPT;
}

/* Null ip_nat_fn: returns NF_ACCEPT. */
unsigned int ip_nat_accept(struct ip_nat_perconn *perconn,
			   struct sk_buff *skb,
			   enum ip_nat_direction direction)
{
	return NF_ACCEPT;
}

/* ip_nat_fn: returns NF_ACCEPT, refreshes timers. */
unsigned int ip_nat_refresh(struct ip_nat_perconn *perconn,
			    struct sk_buff *skb,
			    enum ip_nat_direction dir)
{
	mod_timer(&perconn->timeout, jiffies + perconn->timeend);
	return NF_ACCEPT;
}

static void ipnat_timer_wrapper(unsigned long perconn_ul)
{
	struct ip_nat_perconn *perconn = (struct ip_nat_perconn *)perconn_ul;
	unsigned long flags;

	spin_lock_irqsave(&ip_nat_lock, flags);
	perconn->timerfn(perconn);
	spin_unlock_irqrestore(&ip_nat_lock, flags);
}

/* Returns NULL if can't alloc. */
struct ip_nat_perconn *
ip_nat_alloc_perconn(const struct ip_nat_pktinfo *pkt,
		     const struct ip_nat_binding_element *cause,
		     struct ip_nat_multi_range *range,
		     ip_nat_fn *postmangle, ip_nat_fn *relatedpacket,
		     unsigned long timertime, 
		     void (*timerfn)(struct ip_nat_perconn *perconn))
{
	struct ip_nat_perconn **list;
	struct ip_nat_perconn *perconn;
	const struct ip_nat_tuple_proto *tpr;
	struct ip_nat_multi_range local_range;

	perconn = kmalloc(perconn_size, GFP_ATOMIC);
	if (!perconn) {
		DEBUGP("ip_nat_alloc_perconn: Can't allocate %u bytes\n",
		       perconn_size);
		return NULL;
	}

	perconn->cause = cause;
	perconn->dir[IP_NAT_INCOMING].lasthit 
		= perconn->dir[IP_NAT_OUTGOING].lasthit 
		= jiffies;
	perconn->dir[IP_NAT_INCOMING].hitcount 
		= perconn->dir[IP_NAT_OUTGOING].hitcount
		= 0;
	perconn->postmangle = postmangle;
	perconn->relatedpacket = relatedpacket;
	perconn->dir[IP_NAT_OUTGOING].expired = 0;
	perconn->dir[IP_NAT_INCOMING].expired = 0;

	/* New tuple: start at current IP_NAT_OUTGOING tuple. */
	perconn->allocated_tuple = pkt->tuplepr;
	if (pkt->dir == IP_NAT_INCOMING)
		ip_nat_invert_tuple_proto(&perconn->allocated_tuple);

	/* FIXME: Make this neater -- RR */
	/* 0) If the range is 0-0xFFFFFFFF -> no IP mapping. */
	if (!range->next
	    && range->range.min_ip == 0
	    && range->range.max_ip == 0xFFFFFFFF) {
		u_int32_t ipaddr = (pkt->dir == IP_NAT_OUTGOING 
				    ? pkt->tuplepr.tuple.src.ip
				    : pkt->tuplepr.tuple.dst.ip);
		local_range = ((struct ip_nat_multi_range)
			       { NULL, 0, { ipaddr, ipaddr, 
					    range->range.proto_specific,
					    range->range.min, 
					    range->range.max }});
		range = &local_range;
	}

	if (!get_unique_tuple(&perconn->allocated_tuple, range)) {
		DEBUGP("ip_nat_alloc_perconn: Can't get unique tuple.\n");
		kfree(perconn);
		return NULL;
	}

	dump_tuplepr("alloc_perconn's new OUTGOING perconn", pkt->dir, 
		     &perconn->allocated_tuple);
	if (pkt->dir == IP_NAT_OUTGOING) {
		/* A=Internal src, B=External dest, C=NAT address.
		   Outgoing: A/B -> C/B
		   Incoming: B/C -> B/A

		   tuplepr = A/B, allocated = C/B.
		*/

		/* Outgoing match = A/B.  Outgoing manip = C. */
		perconn->dir[IP_NAT_OUTGOING].connection = pkt->tuplepr;
		perconn->dir[IP_NAT_OUTGOING].manip
			= perconn->allocated_tuple.tuple.src;

		/* Incoming match = B/C
		 * Incoming manip = A. */
		perconn->dir[IP_NAT_INCOMING].connection
			= perconn->allocated_tuple;
		ip_nat_invert_tuple_proto(&perconn->dir[IP_NAT_INCOMING]
					  .connection);
		perconn->dir[IP_NAT_INCOMING].manip = pkt->tuplepr.tuple.src;
	} else {
		/* A=External src, B=Internal dest, C=RNAT address.
		   Incoming: A/B -> A/C
		   Outgoing: C/A -> B/A

		   tuplepr = A/B, allocated = C/A.  */

		/* Incoming match = A/B. */
		perconn->dir[IP_NAT_INCOMING].connection = pkt->tuplepr;
		/* Outgoing match = C/A. */
		perconn->dir[IP_NAT_OUTGOING].connection 
			= perconn->allocated_tuple;

		/* Incoming manip = C */
		perconn->dir[IP_NAT_INCOMING].manip
			= perconn->allocated_tuple.tuple.src;
		/* Outgoing manip = B. */
		{
			struct ip_nat_tuple_proto tmp_tpr = pkt->tuplepr;
			ip_nat_invert_tuple_proto(&tmp_tpr);		
			perconn->dir[IP_NAT_OUTGOING].manip
				= tmp_tpr.tuple.src;
		}
	}

#if 1
	dump_tuplepr("alloc_perconn", IP_NAT_OUTGOING, 
		     &perconn->dir[IP_NAT_OUTGOING].connection);
	dump_manip("alloc_perconn", IP_NAT_OUTGOING, 
		     &perconn->dir[IP_NAT_OUTGOING].manip);
	dump_tuplepr("alloc_perconn", IP_NAT_INCOMING,
		     &perconn->dir[IP_NAT_INCOMING].connection);
	dump_manip("alloc_perconn", IP_NAT_INCOMING, 
		     &perconn->dir[IP_NAT_INCOMING].manip);
#endif

	/* Insert into the various hashes */ 
	list = hash_by_tuple(&perconn->dir[IP_NAT_OUTGOING].connection,
			     IP_NAT_OUTGOING);
	perconn->by_tuple[IP_NAT_OUTGOING] = *list;
	*list = perconn;

	list = hash_by_tuple(&perconn->dir[IP_NAT_INCOMING].connection,
			     IP_NAT_INCOMING);
	perconn->by_tuple[IP_NAT_INCOMING] = *list;
	*list = perconn;

	/* Convenience variable */
	tpr = &perconn->dir[IP_NAT_OUTGOING].connection;
	DEBUGP("ips_proto_hash add src=%u.%u.%u.%u dst=%u.%u.%u.%u proto=%u\n",
	       IP_PARTS(perconn->dir[IP_NAT_OUTGOING].manip.ip), 
	       IP_PARTS(tpr->tuple.dst.ip), tpr->proto);

	list = hash_by_ips_proto(perconn->dir[IP_NAT_OUTGOING].manip.ip, 
				 tpr->tuple.dst.ip, tpr->proto);
	perconn->by_ips_proto = *list;
	*list = perconn;

	list = hash_by_source(tpr);
	perconn->by_source = *list;
	*list = perconn;

	/* Setup expiry timer. */
	perconn->timerfn = timerfn;
	if (timerfn) {
		init_timer(&perconn->timeout);
		DEBUGP("ip_nat_perconn: add_timer %p.\n", 
		       &perconn->timeout);
		perconn->timeout.function = &ipnat_timer_wrapper;
		perconn->timeout.data = (unsigned long)perconn;
		perconn->timeout.expires = jiffies + timertime;
		add_timer(&perconn->timeout);
	}
	perconn->timeend = timertime;

	return perconn;
}

/* Delete all perconns created by a setup. */
static void delete_perconns(const struct ip_nat_binding *bind)
{
	unsigned int i;

	/* We can traverse any of these hashes; chosen at random. */
	for (i = 0; i < IP_NAT_HTABLE_SIZE; i++) {
		struct ip_nat_perconn *perconn;

		for (perconn = conns[0].perconn_by_tuple[IP_NAT_INCOMING][i];
		     perconn;) {
			struct ip_nat_perconn *next 
				= perconn->by_tuple[IP_NAT_INCOMING];

			if (perconn->cause == &bind->elem)
				ip_nat_delete_perconn_timeout(perconn);
			
			perconn = next;
		}
	}
}

#ifdef CONFIG_PROC_FS
static unsigned int dump_dir(char *buffer,
			     const struct ip_nat_perconn *perconn,
			     int dir)
			     
{
	unsigned int len;
	struct ip_nat_proto *proto 
		= find_proto(perconn->dir[dir].connection.proto);
	unsigned long diff, sec, usec;

	diff = jiffies - perconn->dir[dir].lasthit;
	usec = (diff % HZ) * 1000000 / HZ;
	sec = diff / HZ;

	len = sprintf(buffer, "%s:", dir == IP_NAT_INCOMING ? "in" : "out");
	len += sprintf(buffer + len, "%slast=%lu.%06lu hits=%u", 
		       perconn->dir[dir].expired ? "EXPIRED ":"",
		       sec, usec, perconn->dir[dir].hitcount);
	len += sprintf(buffer + len, "src=%u.%u.%u.%u dst=%u.%u.%u.%u ", 
		       IP_PARTS(perconn->dir[dir].connection.tuple.src.ip),
		       IP_PARTS(perconn->dir[dir].connection.tuple.dst.ip));
	len += proto->print_tuple(buffer + len, 
				  &perconn->dir[dir].connection.tuple);
	len += sprintf(buffer + len, "new %s=%u.%u.%u.%u ",
		       (dir == IP_NAT_INCOMING ? "dst" : "src"),
		       IP_PARTS(perconn->dir[dir].manip.ip));

	len += proto->print_manip(buffer + len, &perconn->dir[dir].manip);

	return len;
}

static unsigned int print_perconn(char *buffer,
				  const struct ip_nat_perconn *perconn)
{
	unsigned int len = 0;

	if (perconn->cause) {
		unsigned int i;

		for (i = 0; perconn->cause->type[i]; i++) 
			len += sprintf(buffer + len, "%s:",
				       perconn->cause->type[i]->name);
	}
	else len += sprintf(buffer + len, "null:");

	len += dump_dir(buffer + len, perconn, IP_NAT_INCOMING);
	len += dump_dir(buffer + len, perconn, IP_NAT_OUTGOING);

	if (perconn->cause) {
		unsigned int i;

		for (i = 0; perconn->cause->type[i]; i++) 
			len += perconn->cause->type[i]
				->print_private(buffer + len, perconn);
	}
	len += sprintf(buffer+len, "\n");

	return len;
}

#if 0
static int alloc_null_perconn(const struct iphdr *iph,
			      const struct ip_nat_pktinfo *pkt)
{
	struct ip_nat_perconn *perconn;
	/* We want an exact range. */
	struct ip_nat_multi_range range 
		= { NULL, 0, { pkt->tuplepr.tuple.src.ip,
			       pkt->tuplepr.tuple.src.ip,
			       1,
			       pkt->tuplepr.tuple.src.u,
			       pkt->tuplepr.tuple.src.u } };

	return find_proto(iph->protocol)->bindtype.bind(iph, pkt,
							NULL, &range,
							0, &perconn);
}
#endif

static void cleanup_ip_nat_null_perconns(void)
{
	unsigned int i;
        int local;

	for (local = 0; local < 2; local++)
		/* We can traverse any of these hashes; chosen at random. */
		for (i = 0; i < IP_NAT_HTABLE_SIZE; i++)
			while (conns[local]
			       .perconn_by_tuple[IP_NAT_INCOMING][i] != NULL)
				ip_nat_delete_perconn_timeout(conns[local]
							      .perconn_by_tuple
							      [IP_NAT_INCOMING][i]);
}

/* List of perconns; offset in records not bytes. */
static int nat_get_perconns(char *buffer, char **start, off_t offset,
			    int length, int dummy)
{
	unsigned int i;
	int len = 0,last_len = 0;
	off_t upto = 0;
	unsigned long flags;
	int local = 0;

	spin_lock_irqsave(&ip_nat_lock, flags);
	for (local = 0; local < 2; local++) {
		/* We can traverse any of these hashes; chosen at random. */
		for (i = 0; i < IP_NAT_HTABLE_SIZE; i++) {
			struct ip_nat_perconn *perconn;

			for (perconn = conns[local]
				     .perconn_by_tuple[IP_NAT_INCOMING][i];
			     perconn;
			     perconn = perconn->by_tuple[IP_NAT_INCOMING],
				     upto++) {
				DEBUGP("nat: printing %s entry %lu.\n", 
				       local ? "local" : "non-local", upto);
				if (upto < offset) continue;

				last_len = len;
				len += print_perconn(buffer + len, perconn);
				if (len > length) {
					len = last_len;
					goto finished;
				}
			}
		}
	}

finished:
	spin_unlock_irqrestore(&ip_nat_lock, flags);
	buffer[len] = '\0';

	DEBUGP("nat_get_perconns: Length = %i (of %i).  Offset = %li.\n",
	       len, length, upto);
	/* `start' hack - see fs/proc/generic.c line ~165 */
	*start = (char *)((unsigned int)upto - offset);
	return len;
}
#endif

