#include <linux/types.h>
#include <linux/module.h>
#include <linux/ip.h>
#include <linux/timer.h>
#include <linux/modreglist.h>
#include <net/protocol.h>
#include <net/checksum.h>
#include <linux/bitops.h>
#include <linux/proc_fs.h>
#include <asm/spinlock.h>
#include <linux/netfilter.h>

#include "ip_nat_proto.h"
#include "ip_nat_bind.h"
#include "ip_nat_perconn.h"

#define IP_NAT_TCP_ADD_USER (IP_NAT_SO_MAX+1)
#define IP_NAT_TCP_DELETE_USER (IP_NAT_SO_MAX+2)

#define IP_NAT_TCP_OPENING_TIMEOUT (75*HZ)
#define IP_NAT_TCP_ESTABLISHED_TIMEOUT (6*3600*HZ)
#define IP_NAT_TCP_CLOSED_TIMEOUT (75*HZ)

extern struct ip_nat_proto proto_tcp;

struct ip_nat_tcp_user_port
{
	struct ip_nat_tcp_user_port *next;
	u_int16_t port;
	ip_nat_fn *matchfn;
	ip_nat_fn *relatedfn;
	void (*timerfn)(struct ip_nat_perconn *perconn);
	unsigned long timeout;
};

enum ip_nat_tcpstate {
	IP_NAT_TCP_OPENING,
	IP_NAT_TCP_ESTABLISHED,
	IP_NAT_TCP_CLOSING,
};

static struct ip_nat_tcp_user_port *userports;

/* Use this to track perconn state. */
struct ip_nat_perconn_tcp
{
	enum ip_nat_tcpstate state;

	/* Direction we were initialized in. */
	enum ip_nat_direction bind_dir;
};

static int tcp_check(const struct ip_nat_binding *bind, unsigned int typenum)
{
	/* Must specify proto == TCP. */
	if (bind->elem.mask.tuplepr.proto != 0xFF
	    || bind->elem.info.tuplepr.proto != IPPROTO_TCP) {
		DEBUGP("tcp_bind_check: mask %u, proto = %u.\n",
		       bind->elem.mask.tuplepr.proto,
		       bind->elem.info.tuplepr.proto);
		return 0;
	}

	if (bind->elem.type[typenum+1]) {
		DEBUGP("tcp_bind_check: must be last.\n");
		return 0;
	}

	return 1;
}

/* We don't care when you destroy perconns. */
static void tcp_destroy(const struct ip_nat_perconn *perconn)
{
#ifdef DEBUG
	struct ip_nat_perconn_tcp *perconn_tcp
		= ((void *)perconn) + proto_tcp.bindtype.private_offset;

	if (perconn_tcp->state != IP_NAT_TCP_CLOSING)
		DEBUGP("tcp_destroy: destroying in state %s\n",
		       perconn_tcp->state == IP_NAT_TCP_ESTABLISHED 
		       ? "ESTABLISHED"
		       : (perconn_tcp->state == IP_NAT_TCP_OPENING
			  ? "OPENING" : "UNKNOWN"));
#endif
}

static unsigned int process_icmp(struct ip_nat_perconn *perconn,
				 struct sk_buff *skb,
				 enum ip_nat_direction dir)
{
	struct icmphdr *hdr = (struct icmphdr *)((u_int32_t *)skb->nh.iph 
						 + skb->nh.iph->ihl);
	struct ip_nat_perconn_tcp *perconn_tcp
		= ((void *)perconn) + proto_tcp.bindtype.private_offset;

	if (perconn_tcp->state == IP_NAT_TCP_OPENING
	    && hdr->type == ICMP_DEST_UNREACH) {
		mod_timer(&perconn->timeout, 
			  jiffies + IP_NAT_TCP_CLOSED_TIMEOUT);
		perconn_tcp->state = IP_NAT_TCP_CLOSING;
	}
	return NF_ACCEPT;
}

static unsigned int process(struct ip_nat_perconn *perconn,
			    struct sk_buff *skb,
			    enum ip_nat_direction dir)
{
	const struct iphdr *iph = skb->nh.iph;
	struct tcphdr *hdr = (struct tcphdr *)((u_int32_t *)iph + iph->ihl);
	struct ip_nat_perconn_tcp *perconn_tcp
		= ((void *)perconn) + proto_tcp.bindtype.private_offset;

	switch (perconn_tcp->state) {
	case IP_NAT_TCP_OPENING:
		DEBUGP("ip_nat_tcp: IP_NAT_TCP_OPENING.\n");
		if (dir != perconn_tcp->bind_dir) {
			/* First reply packet: RST -> move to CLOSING,
			 * Other -> move to ESTABLISHED. */
			if (!(ntohs(iph->frag_off) & IP_OFFSET)) {
				if (hdr->rst) {
					mod_timer(&perconn->timeout, 
						  jiffies
						  + IP_NAT_TCP_CLOSED_TIMEOUT);
					DEBUGP("ip_nat_tcp: RST!\n");
					perconn_tcp->state 
						= IP_NAT_TCP_CLOSING;
				} else {
					mod_timer(&perconn->timeout, 
						  jiffies 
						  + IP_NAT_TCP_ESTABLISHED_TIMEOUT);
					DEBUGP("ip_nat_tcp: OK!\n");
					perconn_tcp->state
						= IP_NAT_TCP_ESTABLISHED;
				}
			}
		} else {
			DEBUGP("ip_nat_tcp: Refresh!\n");
			/* Refresh timer. */
			mod_timer(&perconn->timeout, 
				  jiffies + IP_NAT_TCP_OPENING_TIMEOUT);
		}
		break;

	case IP_NAT_TCP_CLOSING:
		DEBUGP("ip_nat_tcp: IP_NAT_TCP_CLOSING.\n");
		/* Nothing can change it now; we're doomed. */
		return NF_ACCEPT;

	case IP_NAT_TCP_ESTABLISHED:
		DEBUGP("ip_nat_tcp: IP_NAT_TCP_ESTABLISHED.\n");
		/* If we see a RST or a FIN, enter expiry cycle. */
		if (!(ntohs(iph->frag_off) & IP_OFFSET) 
		    && (hdr->fin || hdr->rst)) {
			mod_timer(&perconn->timeout, 
				  jiffies + IP_NAT_TCP_CLOSED_TIMEOUT);
			perconn_tcp->state = IP_NAT_TCP_CLOSING;
		} else {
			/* Refresh timer */
			mod_timer(&perconn->timeout, 
				  jiffies + IP_NAT_TCP_ESTABLISHED_TIMEOUT);
		}
		break;
	default:
		DEBUGP("ip_nat_tcp: default %u.\n", perconn_tcp->state);
	}

	return NF_ACCEPT;
}

static struct ip_nat_tcp_user_port no_user_port =
{ NULL, 0, &process, &process_icmp, 
 &ip_nat_delete_perconn_timeout, IP_NAT_TCP_OPENING_TIMEOUT };

static int port_cmp(const void *arg, const void *curr)
{
	return ((struct ip_nat_tcp_user_port *)curr)->port 
		== (u_int16_t)(unsigned long)arg;
}

/* Function to call if we get a match */
static unsigned int tcp_bind(const struct iphdr *iph,
			     const struct ip_nat_pktinfo *pkt,
			     const struct ip_nat_binding_element *cause,
			     struct ip_nat_multi_range *range,
			     unsigned int typenumber,
			     struct ip_nat_perconn **perconn)
{
	struct tcphdr *hdr = (struct tcphdr *)((u_int32_t *)iph + iph->ihl);
	struct ip_nat_tcp_user_port *user;
	struct ip_nat_perconn_tcp *perconn_tcp;

	/* Send to userspace if it's in the user port list. */
	user = modreg_find(&userports, (void *)(int)hdr->dest, &port_cmp);
	if (!user)
		user = &no_user_port;

	dump_tuplepr(__FUNCTION__, pkt->dir, &pkt->tuplepr);
	*perconn = ip_nat_alloc_perconn(pkt, cause, range, user->matchfn,
					user->relatedfn, user->timeout, 
					&ip_nat_delete_perconn_timeout);
	if (!*perconn) {
		DEBUGP("tcp-nat: ip_nat_alloc_perconn returned NULL.\n");
		return NF_DROP;
	}
	else DEBUGP("tcp-nat: ip_nat_alloc_perconn returned ok.\n");

	perconn_tcp = ((void *)*perconn) + proto_tcp.bindtype.private_offset;
	*perconn_tcp = ((struct ip_nat_perconn_tcp)
			{ IP_NAT_TCP_OPENING, pkt->dir });

	return NF_ACCEPT;
}

static int do_tcp_ctl(int cmd, void *user, unsigned int len)
{
	struct ip_nat_tcp_user_port *portinfo;

	switch (cmd) {
	case IP_NAT_TCP_ADD_USER:
		portinfo = kmalloc(sizeof(*portinfo), GFP_USER);
		if (!portinfo)
			return -ENOMEM;
		else {
			portinfo->port = (u_int16_t)(unsigned long)user;
			modreg_prepend(&userports, portinfo);
		}
		break;

	case IP_NAT_TCP_DELETE_USER:
		portinfo = modreg_find(&userports, user, &port_cmp);

		if (!portinfo)
			return -ENOENT;
		else {
			modreg_delete(&userports, portinfo);
			kfree(portinfo);
		}
	}

	return 0;
}

/* Because TCP, UDP and ICMP fields of struct ip_nat_address_range (hangs
   off end) and ip_nat_manip_tuple are structurally equivalent, these
   next two functions are identical for all three, so we cheat and use
   one for all */
static int tcp_in_range(const struct ip_nat_tuple *tuple, 
			const struct ip_nat_range *range)
{
	return !range->proto_specific 
		|| (ntohs(tuple->src.u.tcp.port) >= ntohs(range->min.tcp.port) 
		    && ntohs(tuple->src.u.tcp.port)
		    <= ntohs(range->max.tcp.port));
}

static int tcp_unique_tuple(struct ip_nat_tuple_proto *tuplepr, 
			    const struct ip_nat_range *range)
{
	static u_int16_t port = 0;
	unsigned int range_size;
	struct ip_nat_range newrange;
	unsigned int i;

	/* Try no change, if in range. */
	if (tcp_in_range(&tuplepr->tuple, range)
	    && !ip_nat_used_tuple(tuplepr)) {
		DEBUGP("Straight thru port %hu.\n",
		       ntohs(tuplepr->tuple.src.u.tcp.port));
		return 1;
	}

	/* If no range specified:
	    If it's < 1024, keep it there.
	    Otherwise keep it >= 1024. */
	if (!range->proto_specific) {
		newrange = *range;
		newrange.proto_specific = 1;
		if (tuplepr->tuple.src.u.tcp.port < 1024) {
			DEBUGP("Privileged Range map.\n");
			newrange.min.tcp.port = htons(0);
			newrange.max.tcp.port = htons(1023);
		} else {
			DEBUGP("Unprivileged Range map.\n");
			newrange.min.tcp.port = htons(1024);
			newrange.max.tcp.port = htons(65535);
		}
		range = &newrange;
	}

	range_size = (unsigned int)ntohs(range->max.tcp.port)
		- ntohs(range->min.tcp.port) + 1;

	for (i = 0; i < range_size; i++, port++) {
		tuplepr->tuple.src.u.tcp.port
			= htons(ntohs(range->min.tcp.port) 
				+ (port % range_size));
		if (!ip_nat_used_tuple(tuplepr)) {
			DEBUGP("Allocated port %u.\n", 
			       ntohs(tuplepr->tuple.src.u.tcp.port));
			return 1;
		}
		else DEBUGP("port %u already taken.\n", 
			    ntohs(tuplepr->tuple.src.u.tcp.port));
	}
	return 0;
}

static int tcp_pkt_to_tuple(const void *datah, size_t datalen, 
			    struct ip_nat_tuple *tuple)
{
	const struct tcphdr *hdr = datah;
	if ((char *)&hdr->dest + sizeof(hdr->dest) - (char *)datah > datalen)
		return 0;
	tuple->src.u.tcp.port = hdr->source;
	tuple->dst.u.tcp.port = hdr->dest;

	return 1;
}

static void tcp_manip_pkt(struct iphdr *iph, size_t len,
			  const struct ip_nat_tuple_manip *manip,
			  enum ip_nat_direction direction)
{
	struct tcphdr *hdr = (struct tcphdr *)((u_int32_t *)iph + iph->ihl);
	u_int32_t oldip;
	u_int16_t *portptr;

	if (direction == IP_NAT_OUTGOING) {
		/* Get rid of src ip and src pt */
		oldip = iph->saddr;
		portptr = &hdr->source;
	} else {
		/* Get rid of dst ip and dst pt */
		oldip = iph->daddr;
		portptr = &hdr->dest;
	}
	DEBUGP("TCP: IP %u.%u.%u.%u to %u.%u.%u.%u, port %u to %u\n",
	       IP_PARTS(oldip), IP_PARTS(manip->ip),
	       htons(*portptr), htons(manip->u.tcp.port));

	hdr->check = ip_nat_cheat_check(~oldip, manip->ip,
					ip_nat_cheat_check(*portptr ^ 0xFFFF, 
							   manip->u.tcp.port,
							   hdr->check));
	*portptr = manip->u.tcp.port;
}

static void tcp_invert_tuple(struct ip_nat_tuple *tuple)
{
	/* Switch source and destination ports */
	u_int16_t tmp;

	tmp = tuple->src.u.tcp.port;
	tuple->src.u.tcp.port = tuple->dst.u.tcp.port;
	tuple->dst.u.tcp.port = tmp;
}

static unsigned int 
tcp_print_tuple(char *buffer, const struct ip_nat_tuple *tuple)
{
	return sprintf(buffer, " srcpt=%u dstpt=%u",
		       ntohs(tuple->src.u.tcp.port), 
		       ntohs(tuple->dst.u.tcp.port));
}

static unsigned int
tcp_print_manip(char *buffer, const struct ip_nat_tuple_manip *manip)
{
	return sprintf(buffer, " port=%u", ntohs(manip->u.tcp.port));
}

/* Print out the private part of the perconn (if any). */
static unsigned int tcp_print_private(char *buffer, 
				      const struct ip_nat_perconn *perconn)
{
	struct ip_nat_perconn_tcp *perconn_tcp
		= ((void *)perconn) + proto_tcp.bindtype.private_offset;
	int len = 0;

	switch (perconn_tcp->state) {
	case IP_NAT_TCP_CLOSING:
		len = sprintf(buffer, " CLOSING");
		break;
	case IP_NAT_TCP_OPENING:
		len = sprintf(buffer, " OPENING");
		break;
	case IP_NAT_TCP_ESTABLISHED:
		break;
	}

	/* FIXME: if it's handled by a per-proto, print that here. --RR */
	return len;
}

static struct ip_nat_proto proto_tcp
= { NULL, IPPROTO_TCP, "TCP", 
    tcp_pkt_to_tuple, 
    tcp_manip_pkt, 
    tcp_in_range, 
    tcp_unique_tuple,
    tcp_invert_tuple, 
    tcp_print_tuple, 
    tcp_print_manip, 
    { NULL, "TCP", 0, &__this_module, 
      sizeof(struct ip_nat_perconn_tcp), 0, 
      &tcp_print_private,
      &tcp_check,
      &tcp_bind,
      &tcp_destroy }
};

static struct nf_setsockopt_ops ip_nat_tcp_sockopts 
= { NULL, PF_INET, IP_NAT_TCP_ADD_USER, IP_NAT_TCP_DELETE_USER, do_tcp_ctl };

int init_module(void)
{
	int ret;
	DEBUGP("ip_nat_tcp\n");

	ret = ip_nat_register_proto(&proto_tcp);
	if (ret < 0) return ret;

	ret = nf_register_sockopt(&ip_nat_tcp_sockopts);
	if (ret < 0)
		ip_nat_unregister_proto(&proto_tcp);

	return ret;
}

void cleanup_module()
{
	ip_nat_unregister_proto(&proto_tcp);
	nf_unregister_sockopt(&ip_nat_tcp_sockopts);
}
