/* Program to manage IP Nat binding entries.
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <errno.h>
#include <sys/types.h>
#include <net/if.h>
#include <netinet/in.h>
#include <getopt.h>
#include <string.h>
#include <stdlib.h>
#include <limits.h>
#include <stdarg.h>
#include <netdb.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include "ip_nat.h"
#include "ip_nat_bind.h"

static char version[] = "ipnatctl version 0.1";

#ifndef IPNAT_LIBDIR
#define IPNAT_LIBDIR "/usr/local/lib/ipnatctl"
#endif

#define NEW(type, contents...) 				\
({ 							\
	typeof (type *) _n = malloc(sizeof(type)); 	\
	if (_n) *_n = ( (type) ## contents);		\
	_n;						\
})

void print_usage(char *msg, ...) __attribute__((noreturn));

void print_usage(char *msg, ...)
{
	va_list args;

	va_start(args, msg);
	vfprintf(stderr, msg, args);
	va_end(args);

	fprintf(stderr,
		"\nUsage: "
		"   ipnatctl [-I|-D] <input spec> <output range> <bindings>\n"
		"   ipnatctl -L\n\n");

	fprintf(stderr,
		"<input spec>:\n"
		"  --source/-s [address/mask | address-address]\n"
		"				Source address\n"
		"  --destination/-d [address/mask | address-address]\n"
		"				Destination address\n"
		"  --interface/-i name\n"
		"				Interface to match\n"
		"  --direction/-r [in|out]	Direction of binding\n"
		"  --protocol/-p [name | number] [proto-specifics]\n"
		"				Protocol to match\n\n");

	fprintf(stderr,
		"<output range>:\n"
		"  --to/-t [address/mask | address-address]\n"
		"				IP address range to map to\n\n");

	fprintf(stderr,
		"<bindings>:\n"
		"  --bind <name>[,<name>...]	What to do with matches.\n");

	exit(1);
}

/* FIXME */
int list_bindings(char *argv[])
{
	return system("cat /proc/net/ip_nat_bindings");
}

/* From libops.c */
struct option *
try_load_library(const char *basename,
		 const char *lib, 
		 void (**help)(void), 
		 int (**parse)(int, char **, struct ip_nat_bind_user *),
		 struct option *original_opts);

static struct option original_opts[] = {
	{ "source", 1, 0, 's', },
	{ "destination", 1, 0, 'd' },
	{ "interface", 1, 0, 'i' },
	{ "protocol", 1, 0, 'p' },
	{ "direction", 1, 0, 'r' },
	{ "to", 1, 0, 't' },
	{ "binding", 1, 0, 'b' },
	{ "help", 0, 0, 'h' },
	{ "version", 0, 0, 'v' },
	{ NULL, 0, 0, 0 }
};

static int get_address(const char *string, u_int32_t *ip, u_int32_t *mask)
{
	struct in_addr addr;
	char *delim;

	delim = strchr(string, '/');
	if (delim) {
		struct in_addr addrmask;

		if (strchr(delim+1, '.')) {
			if (!inet_aton(delim+1, &addrmask))
				return 0;
			*mask = addrmask.s_addr;

			*delim = '\0';
			if (!inet_aton(string, &addr))
				return 0;
			*ip = addr.s_addr & *mask;

			return 1;
		}
		else {
			unsigned long tmpmask;

			errno = 0;
			tmpmask = strtoul(delim+1, NULL, 10);
			if (tmpmask == ULONG_MAX && errno)
				return 0;

			*mask = htonl((~0) << (32 - tmpmask));
			*delim = '\0';
			if (!inet_aton(string, &addr))
				return 0;

			*ip = addr.s_addr & *mask;

			return 1;
		}
	}
	else {
		if (!inet_aton(string, &addr))
			return 0;
		*ip = addr.s_addr;
	}

	*mask = ~0;
	return 1;
}

static void
get_source_address(struct ip_nat_bind_user *base, const char *optarg)
{
	if (!get_address(optarg, &base->tupleproto.tuple.src.ip,
			 &base->mask.tuplepr.tuple.src.ip))
		print_usage("Bad source address `%s'", optarg);
}

static void
get_destination_address(struct ip_nat_bind_user *base,
			const char *optarg)
{
	if (!get_address(optarg, &base->tupleproto.tuple.dst.ip,
			 &base->mask.tuplepr.tuple.dst.ip))
		print_usage("Bad destination address `%s'", optarg);
}

static void
get_interface(struct ip_nat_bind_user *base, const char *optarg)
{
	if (strlen(optarg) > IFNAMSIZ-1)
		print_usage("Interface name `%s' too long", optarg);
	strcpy(base->ifname, optarg);
	memset(&base->mask.device, 0xFF, sizeof(base->mask.device));
}

/* In case no /etc/protocols. */
struct pprot {
	char *name;
	unsigned short num;
};

static const struct pprot default_protos[] = {
	{ "all", 0 },
	{ "tcp", IPPROTO_TCP },
	{ "udp", IPPROTO_UDP },
	{ "icmp", IPPROTO_ICMP }
};

static void
get_protocol(struct ip_nat_bind_user *base, const char *optarg,
	     struct option **opts, 
	     void (**help)(void), 
	     int (**parse)(int, char **, struct ip_nat_bind_user *))
{
	int proto = atoi(optarg);
	char protoname[5];

	if (proto < 1 || proto > 255) {
		struct protoent *pent;

		pent = getprotobyname(optarg);
		if (pent)
			proto = pent->p_proto;
		else {
			unsigned int i;

			for (i = 0; 
			     i < sizeof(default_protos)/sizeof(struct pprot);
			     i++) {
				if (strcasecmp(optarg, default_protos[i].name)
				    == 0)
					proto = default_protos[i].num;
			}

			if (i == sizeof(default_protos)/sizeof(struct pprot)) {
				print_usage("unknown protocol `%s' specified",
					    optarg);
			}
		}
	}

	base->tupleproto.proto = proto;
	base->mask.tuplepr.proto = 0xFF;

	sprintf(protoname, "%i", proto);
	*opts = try_load_library(IPNAT_LIBDIR "/libnatctl_proto_", protoname,
				 help, parse, *opts);
}

static void
get_direction(struct ip_nat_bind_user *base, const char *optarg)
{
	if (strncasecmp(optarg, "input", strlen(optarg) ?: 1) == 0)
		base->dir = IP_NAT_INCOMING;
	else if (strncasecmp(optarg, "output", strlen(optarg) ?: 1) == 0)
		base->dir = IP_NAT_OUTGOING;
	else 
		print_usage("Unknown direction `%s'", optarg);
}

static int 
get_address_range(const char *string, u_int32_t *ipmin, u_int32_t *ipmax)
{
	struct in_addr minaddr;
	char *delim;

	delim = strchr(string, ':');
	if (delim) {
		struct in_addr maxaddr;

		if (!inet_aton(delim+1, &maxaddr))
			return 0;
		*ipmax = maxaddr.s_addr;
	}
	else if ((delim = strchr(string, '/')) != NULL) {
		if (strchr(delim+1, '.')) {
			struct in_addr addrmask;
			if (!inet_aton(delim+1, &addrmask))
				return 0;

			*delim = '\0';
			if (inet_aton(string, &minaddr) == 0)
				return 0;

			*ipmin = (minaddr.s_addr & addrmask.s_addr);
			*ipmax = (minaddr.s_addr | ~addrmask.s_addr);
		} else {
			unsigned long mask;
			errno = 0;
			mask = strtoul(delim+1, NULL, 10);
			if (mask == ULONG_MAX && errno)
				return 0;

			mask = htonl((~0) << (32 - mask));
			*delim = '\0';
			if (inet_aton(string, &minaddr) == 0)
				return 0;

			*ipmin = (minaddr.s_addr & mask);
			*ipmax = (minaddr.s_addr | ~mask);
		}
	} else {
		if (!inet_aton(string, &minaddr))
			return 0;
		*ipmax = *ipmin = minaddr.s_addr;
	}

	return (ntohl(*ipmax) >= ntohl(*ipmin));
}

static void
get_to(struct ip_nat_bind_user *base, const char *optarg)
{
	if (!get_address_range(optarg, 
			       &base->range.min_ip, &base->range.max_ip)) 
		print_usage("Bad address range `%s'", optarg);
}

static void 
get_bindings(struct ip_nat_bind_user **base, 
	     char *optarg, 
	     struct ip_nat_bind_user *bind, 
	     size_t *size)
{
	char *name;
	unsigned int i = 0;

	*size = sizeof(**base) + strlen(optarg) + 2;
	*base = realloc(*base, *size);

	name = strtok(optarg, ",");
	if (!name)
		print_usage("Expected a binding name");
		
	while (name) {
		strcpy((*base)->typenames+i, name);
		i += strlen(name) + 1;

		name = strtok(NULL, ",");
	}
	/* Double-nul terminator. */
	(*base)->typenames[i] = '\0';

	return;
}

enum
{
	SPECIFIED_DIRECTION = 1,
	SPECIFIED_BINDING = 2
};

static struct ip_nat_bind_user *parse(int argc, char *argv[], size_t *size)
{
	struct ip_nat_bind_user *bind 
		= NEW(struct ip_nat_bind_user, { 0 });
	struct option *opts = original_opts;
	int (*parse)(int, char **, struct ip_nat_bind_user *) = NULL;
	void (*help)(void) = NULL;
	int c;
	unsigned int option_flags = 0;

	*size = sizeof(struct ip_nat_bind_user);

	/* FIXME: Set --to range to all; is this reliable for
	   -p tcp -i plip0 -r in --dport 80 --to-port 25 
	   say? --RR */
	bind->range = ((struct ip_nat_range) { 0, 0xFFFFFFFF, 0 });

	while ((c = getopt_long(argc, argv, "s:d:i:p:r:t:b:", opts, NULL)) 
	       != -1) {
		switch (c) {
		case 's':
			get_source_address(bind, optarg);
			break;

		case 'd':
			get_destination_address(bind, optarg);
			break;

		case 'i':
			get_interface(bind, optarg);
			break;

		case 'p':
			get_protocol(bind, optarg, &opts, &help, &parse);
			break;

		case 'r':
			get_direction(bind, optarg);
			option_flags |= SPECIFIED_DIRECTION;
			break;

		case 't':
			get_to(bind, optarg);
			break;

		case 'b':
			get_bindings(&bind, optarg, bind, size);
			option_flags |= SPECIFIED_BINDING;
			break;

		case 'h':
			if (help) {
				help();
				exit(0);
			} else
				print_usage("ipnatctl help\n");

		case 'v':
			printf(version);
			exit(0);

		default:
			if (!parse || !parse(c, argv, bind))
				print_usage("Unknown argument `%s'", optarg);
		}
	}

	if (optind < argc)
		print_usage("unknown arguments found on commandline");

	if (option_flags != (SPECIFIED_DIRECTION + SPECIFIED_BINDING))
		print_usage("Need --binding/-b and --direction/-r options.");

	return bind;
}

int main(int argc, char *argv[])
{
	int sockfd, op;
	size_t size;
	static struct ip_nat_bind_user *bind;

	if (argc < 2)
		print_usage("Too few options.");

	if (strcmp(argv[1], "-I") == 0)
		op = IP_NAT_SO_INSERT;
	else if (strcmp(argv[1], "-D") == 0)
		op = IP_NAT_SO_DELETE;
	else if (strcmp(argv[1], "-L") == 0)
		return list_bindings(argv+1);
	else print_usage("Unknown option `%s'", argv[1]);

	bind = parse(argc-1, argv+1, &size);

#if 1
	fprintf(stderr, "Range for --to: %u.%u.%u.%u - %u.%u.%u.%u.\n",
		ntohl(bind->range.min_ip) >> 24,
		(ntohl(bind->range.min_ip) >> 16) & 0xFF,
		(ntohl(bind->range.min_ip) >> 8) & 0xFF,
		ntohl(bind->range.min_ip) & 0xFF,
		ntohl(bind->range.max_ip) >> 24,
		(ntohl(bind->range.max_ip) >> 16) & 0xFF,
		(ntohl(bind->range.max_ip) >> 8) & 0xFF,
		ntohl(bind->range.max_ip) & 0xFF);

	fprintf(stderr, "Proto-specific: %i.\n", bind->range.proto_specific);
	if (bind->range.proto_specific)
		fprintf(stderr, "Ports %u to %u.\n",
			ntohs(bind->range.min.tcp.port),
			ntohs(bind->range.max.tcp.port));
#endif

	sockfd = socket(AF_INET, SOCK_RAW, IPPROTO_RAW);
	if (sockfd < 0) {
		perror("Opening raw socket");
		exit(1);
	}

	if (setsockopt(sockfd, IPPROTO_IP, op, bind, size) < 0) {
		perror("setsockopt problem");
		return 1;
	}
	return 0;
}
